\ Xatom.4th
\
\
\  Do not change!  Changing this file implies a protocol change!
\
 1  constant  XA_PRIMARY
 2  constant  XA_SECONDARY
 3  constant  XA_ARC
 4  constant  XA_ATOM
 5  constant  XA_BITMAP
 6  constant  XA_CARDINAL
 7  constant  XA_COLORMAP
 8  constant  XA_CURSOR
 9  constant  XA_CUT_BUFFER0
10  constant  XA_CUT_BUFFER1
11  constant  XA_CUT_BUFFER2
12  constant  XA_CUT_BUFFER3
13  constant  XA_CUT_BUFFER4
14  constant  XA_CUT_BUFFER5
15  constant  XA_CUT_BUFFER6
16  constant  XA_CUT_BUFFER7
17  constant  XA_DRAWABLE
18  constant  XA_FONT
19  constant  XA_INTEGER
20  constant  XA_PIXMAP
21  constant  XA_POINT
22  constant  XA_RECTANGLE
23  constant  XA_RESOURCE_MANAGER
24  constant  XA_RGB_COLOR_MAP
25  constant  XA_RGB_BEST_MAP
26  constant  XA_RGB_BLUE_MAP
27  constant  XA_RGB_DEFAULT_MAP
28  constant  XA_RGB_GRAY_MAP
29  constant  XA_RGB_GREEN_MAP
30  constant  XA_RGB_RED_MAP
31  constant  XA_STRING
32  constant  XA_VISUALID
33  constant  XA_WINDOW
34  constant  XA_WM_COMMAND
35  constant  XA_WM_HINTS
36  constant  XA_WM_CLIENT_MACHINE
37  constant  XA_WM_ICON_NAME
38  constant  XA_WM_ICON_SIZE
39  constant  XA_WM_NAME
40  constant  XA_WM_NORMAL_HINTS
41  constant  XA_WM_SIZE_HINTS
42  constant  XA_WM_ZOOM_HINTS
43  constant  XA_MIN_SPACE
44  constant  XA_NORM_SPACE
45  constant  XA_MAX_SPACE
46  constant  XA_END_SPACE
47  constant  XA_SUPERSCRIPT_X
48  constant  XA_SUPERSCRIPT_Y
49  constant  XA_SUBSCRIPT_X
50  constant  XA_SUBSCRIPT_Y
51  constant  XA_UNDERLINE_POSITION
52  constant  XA_UNDERLINE_THICKNESS
53  constant  XA_STRIKEOUT_ASCENT
54  constant  XA_STRIKEOUT_DESCENT 
55  constant  XA_ITALIC_ANGLE
56  constant  XA_X_HEIGHT
57  constant  XA_QUAD_WIDTH
58  constant  XA_WEIGHT
59  constant  XA_POINT_SIZE
60  constant  XA_RESOLUTION
61  constant  XA_COPYRIGHT
62  constant  XA_NOTICE
63  constant  XA_FONT_NAME
64  constant  XA_FAMILY_NAME
65  constant  XA_FULL_NAME
66  constant  XA_CAP_HEIGHT
67  constant  XA_WM_CLASS
68  constant  XA_WM_TRANSIENT_FOR

68  constant  XA_LAST_PREDEFINED

